#p_win {
}

#c_win, #c_frame{
    background-color: #000;
}

#p_win GtkPaned .pane-separator {
    background-color: #ccc;
    border-radius: 2px;
}

#p_win GtkFrame {
}

#p_win GtkTreeView {
    border-style: none;
    background-color: transparent;
    font-size: 16pt;
}

#scribble_overlay {
  background: @theme_bg_color
}

.frame-label {
    font-size:1.2em;
}

.big-info-label {
    font-size: 1.6em;
}

.info-label {
    font-size: 1.4em;
    opacity: 0.80;
}

.big-info-label > .frame-label {
    font-size:0.7em;
}

.info-label > .frame-label {
    font-size:0.8em;
}

.ett-reached {
    color: #64c864;
}
.ett-info {
    color: #FFBF00;
}
.ett-warn {
    color: #FF0000;
}
.time-warn {
    animation: blinker 250ms cubic-bezier(.5, 0, 1, 1) infinite alternate;
}
@keyframes blinker {
  from { opacity: 1; }
  to { opacity: 0; }
}

#MenuBar,
#MenuBar * {
}
#MenuBar *:hover {
}
