@import url("default.css");

#p_win {
    background-color: #444;
}

#p_win GtkPaned .pane-separator {
    background-color: #aaa;
}

#p_win GtkTreeView {
    color: #aaa;
}

.frame-label {
    color: #eee;
    padding-bottom: 5px;
    font-size: 110%;
}

.info-label {
    color: #aaa;
    font-size: 32pt;
}

.ett-info {
    color: #FFEA09;
}

#MenuBar,
#MenuBar GtkMenuItem {
    color: #eee;
    background-color: #aaa;
}
#MenuBar *:hover {
    color: #eee;
    background-color: #444;
}
